using System;
using System.Text;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Engines;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Crypto.Modes;
using Org.BouncyCastle.Utilities.IO;


public class AeadCipherExamples
{
 
    public static void Main(String[] args)
    {
        
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("===============================");
        Console.WriteLine("Start Aead Cipher Examples.Main");
        Console.WriteLine("===============================");


        {
            //Example 15 – GCM AEAD Cipher
            Console.WriteLine();
            Console.WriteLine("*** Example 15 – GCM AEAD Cipher ***");
            ExValues.cSharpFixedRandom.SetSeed(151);
            KeyParameter keyParam = keyParameterGeneration(128);
            
            byte[] cipherTextData =  gcmAEADEncrypt(keyParam, 
                ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" GCM AEAD Cipher")).ToArray());
            Console.WriteLine(string.Format("Encrypt GCM AEAD Cipher\tplainTextData:[{0}]\tLength:[{1}]\tcipherTextData:[{2}]\tLength:[{3}]",
                Encoding.Default.GetString(ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" GCM AEAD Cipher")).ToArray()),
                Encoding.Default.GetString(ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" GCM AEAD Cipher")).ToArray()).Length,
                string.Join(",",cipherTextData),
                cipherTextData.Length));
            byte[] plainTextData = gcmAEADDecrypt(keyParam, cipherTextData);
            Console.WriteLine(string.Format("Decrypt GCM AEAD Cipher\tcipherTextData:[{0}]\tLength:[{1}]\tplainTextData:[{2}]\tLength:[{3}]",
                string.Join(",",cipherTextData),
                cipherTextData.Length.ToString(),
                Encoding.Default.GetString(plainTextData),
                Encoding.Default.GetString(plainTextData).Length
                ));
            
       }        
       
        {
            //Example 16 – ASCON AEAD Cipher
            Console.WriteLine();
            Console.WriteLine("*** Example 16 – ASCON AEAD Cipher ***");
            ExValues.cSharpFixedRandom.SetSeed(151);
            KeyParameter keyParam = keyParameterGeneration(128);
            
            byte[] cipherTextData =  asconAEADEncrypt(keyParam, 
                ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" ASCON AEAD Cipher")).ToArray());
            Console.WriteLine(string.Format("Encrypt ASCON AEAD Cipher\tplainTextData:[{0}]\tLength:[{1}]\tcipherTextData:[{2}]\tLength:[{3}]",
                Encoding.Default.GetString(ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" ASCON AEAD Cipher")).ToArray()),
                Encoding.Default.GetString(ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" ASCON AEAD Cipher")).ToArray()).Length,
                string.Join(",",cipherTextData),
                cipherTextData.Length));
            byte[] plainTextData = asconAEADDecrypt(keyParam, cipherTextData);
            Console.WriteLine(string.Format("Decrypt ASCON AEAD Cipher\tcipherTextData:[{0}]\tLength:[{1}]\tplainTextData:[{2}]\tLength:[{3}]",
                string.Join(",",cipherTextData),
                cipherTextData.Length.ToString(),
                Encoding.Default.GetString(plainTextData),
                Encoding.Default.GetString(plainTextData).Length
                ));
            
       }        


        {
            //Example 17 – ChaCha20Poly1305 AEAD Cipher
            Console.WriteLine();
            Console.WriteLine("*** Example 17 – ChaCha20Poly1305 AEAD Cipher ***");
            ExValues.cSharpFixedRandom.SetSeed(151);
            KeyParameter keyParam = keyParameterGeneration(256);
            
            byte[] cipherTextData =  chacha20poly1305AEADEncrypt(keyParam, 
                ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" ChaCha20Poly1305 AEAD Cipher")).ToArray());
            Console.WriteLine(string.Format("Encrypt ChaCha20Poly1305 AEAD Cipher\tplainTextData:[{0}]\tLength:[{1}]\tcipherTextData:[{2}]\tLength:[{3}]",
                Encoding.Default.GetString(ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" ChaCha20Poly1305 AEAD Cipher")).ToArray()),
                Encoding.Default.GetString(ExValues.sampleInput.Concat(Encoding.Default.GetBytes(" ChaCha20Poly1305 AEAD Cipher")).ToArray()).Length,
                string.Join(",",cipherTextData),
                cipherTextData.Length));
            byte[] plainTextData = chacha20poly1305AEADDecrypt(keyParam, cipherTextData);
            Console.WriteLine(string.Format("Decrypt ChaCha20Poly1305 AEAD Cipher\tcipherTextData:[{0}]\tLength:[{1}]\tplainTextData:[{2}]\tLength:[{3}]",
                string.Join(",",cipherTextData),
                cipherTextData.Length.ToString(),
                Encoding.Default.GetString(plainTextData),
                Encoding.Default.GetString(plainTextData).Length
                ));
            
       }        
       
        Console.WriteLine("===============================");
        Console.WriteLine("End   Aead Cipher Examples.Main");
        Console.WriteLine("===============================");
        Console.WriteLine();
        Console.WriteLine();        
    }
    
    /// <summary>
    /// Generate appropriate random key 
    /// </summary>        
    public static KeyParameter keyParameterGeneration(int keySize)
    {
        CipherKeyGenerator keyGen = new CipherKeyGenerator();        
        keyGen.Init(new KeyGenerationParameters(ExValues.cSharpFixedRandom, keySize));
        KeyParameter keyParam = keyGen.GenerateKeyParameter();
        return keyParam;
    }

    /// <summary>
    /// Generate key from a given set of bytes
    /// </summary>        
    public static KeyParameter keyParameterGeneration(byte[] myKey)
    {
        KeyParameter keyParam = new KeyParameter(myKey);        
        return keyParam;
    }
    
    /// <summary>
    /// Key and Initialisation Vector Generation
    /// </summary>        
    public static ParametersWithIV keyParameterGenerationWithIV(int keySize, byte[] IV)
    {
        CipherKeyGenerator keyGen = new CipherKeyGenerator();        
        keyGen.Init(new KeyGenerationParameters(ExValues.cSharpFixedRandom, keySize));
        KeyParameter keyParam = keyGen.GenerateKeyParameter();
        ParametersWithIV keyParamWithIV = new ParametersWithIV(keyParam, IV);
        return keyParamWithIV;
    }

    /// <summary>
    /// Key, Initialisation Vector and Associated Data Generation
    /// </summary>        
    public static AeadParameters aeadParameterGeneration(int keySize, int macSize, byte[] IV, byte[] associatedText)
    {
        CipherKeyGenerator keyGen = new CipherKeyGenerator();        
        keyGen.Init(new KeyGenerationParameters(ExValues.cSharpFixedRandom, keySize));
        KeyParameter keyParam = keyGen.GenerateKeyParameter();
        AeadParameters aeadKeyParam = new AeadParameters(keyParam, macSize, IV, associatedText);
        return aeadKeyParam;
    }

    /// <summary>
    /// Example 15 – GCM AEAD Mode Encryption
    /// </summary>        
    public static byte[] gcmAEADEncrypt(KeyParameter keyParam, byte[] plainTextData)
    {
        IBlockCipher cipher = new AesEngine();        
        int macSize = 8*cipher.GetBlockSize();
        byte[] nonce = ExValues.sampleIVnonce;
        byte[] associatedText = ExValues.additionalAuthenticatedDataA;
        AeadParameters keyParamAead = new AeadParameters(keyParam, macSize, nonce, associatedText);

        GcmBlockCipher cipherMode = new GcmBlockCipher(cipher);        
        cipherMode.Init(true,keyParamAead);
        
        int outputSize = cipherMode.GetOutputSize(plainTextData.Length);
        byte[]  cipherTextData = new byte[outputSize];
        int result = cipherMode.ProcessBytes(plainTextData, 0, plainTextData.Length, cipherTextData, 0);
        cipherMode.DoFinal(cipherTextData, result);
        return cipherTextData;
    }

    /// <summary>
    /// Example 15 – GCM AEAD Mode Decryption
    /// </summary>        
    public static byte[] gcmAEADDecrypt(KeyParameter keyParam, byte[] cipherTextData)
    {
        IBlockCipher cipher = new AesEngine();        
        int macSize = 8*cipher.GetBlockSize();
        byte[] nonce = ExValues.sampleIVnonce;
        byte[] associatedText = ExValues.additionalAuthenticatedDataA;
        AeadParameters keyParamAead = new AeadParameters(keyParam, macSize, nonce, associatedText);

        GcmBlockCipher cipherMode = new GcmBlockCipher(cipher);        
        cipherMode.Init(false,keyParamAead);
        
        int outputSize = cipherMode.GetOutputSize(cipherTextData.Length);
        byte[]  plainTextData = new byte[outputSize];
        int result = cipherMode.ProcessBytes(cipherTextData, 0, cipherTextData.Length, plainTextData, 0);
        
        cipherMode.DoFinal(plainTextData, result);
        return plainTextData;
    }

    /// <summary>
    /// Example 16 – ASCON AEAD Encryption
    /// </summary>        
    public static byte[] asconAEADEncrypt(KeyParameter keyParam, byte[] plainTextData)
    {
        AsconEngine cipher = new AsconEngine(AsconEngine.AsconParameters.ascon128a);        
        int macSize = 8*cipher.GetKeyBytesSize();
        byte[] nonce = new byte[cipher.GetIVBytesSize()];
        Array.Copy(ExValues.sampleIVnonce, nonce, cipher.GetIVBytesSize());
        byte[] associatedText = ExValues.additionalAuthenticatedDataA;
        AeadParameters keyParamAead = new AeadParameters(keyParam, macSize,nonce, associatedText);            
        cipher.Init(true,keyParamAead);
        int outputSize = cipher.GetOutputSize(plainTextData.Length);
        byte[]  cipherTextData = new byte[outputSize];        
        int result = cipher.ProcessBytes(plainTextData, 0, plainTextData.Length, cipherTextData, 0);
        cipher.DoFinal(cipherTextData,result);
        return cipherTextData;
    }
  
    /// <summary>
    /// Example 16 – ASCON AEAD Decryption
    /// </summary>        
    public static byte[] asconAEADDecrypt(KeyParameter keyParam, byte[] cipherTextData)
    {
        AsconEngine cipher = new AsconEngine(AsconEngine.AsconParameters.ascon128a);        
        int macSize = 8*cipher.GetKeyBytesSize();
        byte[] nonce = new byte[cipher.GetIVBytesSize()];
        Array.Copy(ExValues.sampleIVnonce, nonce, cipher.GetIVBytesSize());
        byte[] associatedText = ExValues.additionalAuthenticatedDataA;
        AeadParameters keyParamAead = new AeadParameters(keyParam, macSize,nonce, associatedText);            
        cipher.Init(false,keyParamAead);
        int outputSize = cipher.GetOutputSize(cipherTextData.Length);
        byte[]  plainTextData = new byte[outputSize];        
        int result = cipher.ProcessBytes(cipherTextData, 0, cipherTextData.Length, plainTextData, 0);
        cipher.DoFinal(plainTextData,result);
        return plainTextData;
    }   

    /// <summary>
    /// Example 17 – ChaCha20Poly1305 AEAD Encryption
    /// </summary>        
    public static byte[] chacha20poly1305AEADEncrypt(KeyParameter keyParam, byte[] plainTextData)
    {
        ChaCha20Poly1305 cipher = new ChaCha20Poly1305();        
        int macSize = 128;
        byte[] nonce = new byte[12];
        Array.Copy(ExValues.sampleIVnonce, nonce, 12);
        byte[] associatedText = ExValues.additionalAuthenticatedDataA;
        AeadParameters keyParamAead = new AeadParameters(keyParam, macSize,nonce, associatedText);            
        
        cipher.Init(true,keyParamAead);
        
        int outputSize = cipher.GetOutputSize(plainTextData.Length);
        byte[]  cipherTextData = new byte[outputSize];        
        int result = cipher.ProcessBytes(plainTextData, 0, plainTextData.Length, cipherTextData, 0);
        cipher.DoFinal(cipherTextData,result);
        return cipherTextData;
    }
  
    /// <summary>
    /// Example 17 – ChaCha20Poly1305 AEAD Decryption
    /// </summary>        
    public static byte[] chacha20poly1305AEADDecrypt(KeyParameter keyParam, byte[] cipherTextData)
    {
        ChaCha20Poly1305 cipher = new ChaCha20Poly1305();        
        int macSize = 128;
        byte[] nonce = new byte[12];
        Array.Copy(ExValues.sampleIVnonce, nonce, 12);
        byte[] associatedText = ExValues.additionalAuthenticatedDataA;
        AeadParameters keyParamAead = new AeadParameters(keyParam, macSize,nonce, associatedText);            

        cipher.Init(false,keyParamAead);

        int outputSize = cipher.GetOutputSize(cipherTextData.Length);
        byte[]  plainTextData = new byte[outputSize];        
        int result = cipher.ProcessBytes(cipherTextData, 0, cipherTextData.Length, plainTextData, 0);
        cipher.DoFinal(plainTextData,result);
        return plainTextData;
    }       
}
