using System;
using System.Text;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Engines;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Crypto.Paddings;
using Org.BouncyCastle.Utilities;
using Org.BouncyCastle.Utilities.Encoders;
using Org.BouncyCastle.Crypto.Utilities;
using Org.BouncyCastle.Crypto.Modes;
using Org.BouncyCastle.Utilities.IO;


public class KeyWrapExamples
{
 
    public static void Main(String[] args)
    {
        
        Console.WriteLine();
        Console.WriteLine();
        Console.WriteLine("================================");
        Console.WriteLine("Start Key Wrapping Examples.Main");
        Console.WriteLine("================================");


        {
            // Example 29 – AES Key Wrapping without padding
            Console.WriteLine();
            Console.WriteLine("*** Example 29 – AES Key Wrapping without padding ***");
            ExValues.cSharpFixedRandom.SetSeed(151);
            ICipherParameters keyParam = SymmetricCipherExamples.keyParameterGeneration(256);

            byte[] wrappedData = wrapKeyRFC3394(keyParam, ExValues.sampleUnpaddedKey);
            Console.WriteLine(string.Format("Wrapped Data \t[{0}]\tLength:[{1}]\t[{2}]\tLength:[{3}]",
                string.Join(",",wrappedData),
                wrappedData.Length,
                Encoding.Default.GetString(ExValues.sampleUnpaddedKey),
                Encoding.Default.GetString(ExValues.sampleUnpaddedKey).Length));
                
            byte[] unwrappedData = unwrapKeyRFC3394(keyParam, wrappedData);
            Console.WriteLine(string.Format("UnWrapped Data \t[{0}]\tLength:[{1}]\t[{2}]\tLength:[{3}]",
                Encoding.Default.GetString(unwrappedData),
                Encoding.Default.GetString(unwrappedData).Length,
                Encoding.Default.GetString(ExValues.sampleUnpaddedKey),
                Encoding.Default.GetString(ExValues.sampleUnpaddedKey).Length));
        }

        {
            ExValues.cSharpFixedRandom.SetSeed(151);
            ICipherParameters keyParam = SymmetricCipherExamples.keyParameterGeneration(256);

            byte[] wrappedData = wrapKeyRFC3394WithIV(keyParam, ExValues.sampleShortIVnonce, ExValues.sampleUnpaddedKey);
            Console.WriteLine(string.Format("Wrapped Data \t[{0}]\tLength:[{1}]\t[{2}]\tLength:[{3}]",
                string.Join(",",wrappedData),
                wrappedData.Length,
                Encoding.Default.GetString(ExValues.sampleUnpaddedKey),
                Encoding.Default.GetString(ExValues.sampleUnpaddedKey).Length));
                
            byte[] unwrappedData = unwrapKeyRFC3394WithIV(keyParam, ExValues.sampleShortIVnonce, wrappedData);
            Console.WriteLine(string.Format("UnWrapped Data \t[{0}]\tLength:[{1}]\t[{2}]\tLength:[{3}]",
                Encoding.Default.GetString(unwrappedData),
                Encoding.Default.GetString(unwrappedData).Length,
                Encoding.Default.GetString(ExValues.sampleUnpaddedKey),
                Encoding.Default.GetString(ExValues.sampleUnpaddedKey).Length));
        }

        Console.WriteLine("================================");
        Console.WriteLine("End   Key Wrapping Examples.Main");
        Console.WriteLine("================================");
        Console.WriteLine();
        Console.WriteLine();        
    }

    public static byte[] wrapKeyRFC3394(ICipherParameters keyParam, byte[] unwrappedData)
    {
        IBlockCipher symmetricBlockCipher = new AesEngine();  
        Rfc3394WrapEngine wrapEngine = new Rfc3394WrapEngine(symmetricBlockCipher);
        wrapEngine.Init(true, keyParam);
        return wrapEngine.Wrap(unwrappedData,0,unwrappedData.Length);
    }

    public static byte[] unwrapKeyRFC3394(ICipherParameters keyParam, byte[] wrappedData)
    {
        IBlockCipher symmetricBlockCipher = new AesEngine();  
        Rfc3394WrapEngine wrapEngine = new Rfc3394WrapEngine(symmetricBlockCipher);
        wrapEngine.Init(false, keyParam);
        return wrapEngine.Unwrap(wrappedData,0,wrappedData.Length);
    }
    
    public static byte[] wrapKeyRFC3394WithIV(ICipherParameters keyParam, byte[] IV, byte[] unwrappedData)
    {
        ParametersWithIV keyParamWithIV = new ParametersWithIV(keyParam, IV);

        IBlockCipher symmetricBlockCipher = new AesEngine();  
        Rfc3394WrapEngine wrapEngine = new Rfc3394WrapEngine(symmetricBlockCipher);
        wrapEngine.Init(true, keyParamWithIV);
        return wrapEngine.Wrap(unwrappedData,0,unwrappedData.Length);
    }

    public static byte[] unwrapKeyRFC3394WithIV(ICipherParameters keyParam, byte[] IV, byte[] wrappedData)
    {
        ParametersWithIV keyParamWithIV = new ParametersWithIV(keyParam, IV);

        IBlockCipher symmetricBlockCipher = new AesEngine();  
        Rfc3394WrapEngine wrapEngine = new Rfc3394WrapEngine(symmetricBlockCipher);
        wrapEngine.Init(false, keyParamWithIV);
        return wrapEngine.Unwrap(wrappedData,0,wrappedData.Length);
    }
    
}
